"""Class representing errors encountered during the reading of the calibration file.

**Authors**: SDU
"""


class CalibrationError(Exception):
    """Class representing errors encountered during the reading of the calibration."""
